@@ -1,5 +1,5 @@
 /* 
- * Copyright (C) 2006, 2007 OpenedHand Ltd.
+ * Copyright (C) 2006, 2007, 2008 OpenedHand Ltd.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
  *
@@ -31,7 +31,7 @@
  */
 
 #include <config.h>
-#include <libsoup/soup-date.h>
+#include <libsoup/soup.h>
 #include <string.h>
 #include <stdio.h>
 
@@ -40,6 +40,9 @@
 #include "gssdp-protocol.h"
 #include "gssdp-marshal.h"
 
+#define MAX_DISCOVERY_MESSAGES 3
+#define DISCOVERY_FREQUENCY    500 /* 500 ms */
+
 G_DEFINE_TYPE (GSSDPResourceBrowser,
                gssdp_resource_browser,
                G_TYPE_OBJECT);
@@ -56,6 +59,9 @@ struct _GSSDPResourceBrowserPrivate {
         gulong       message_received_id;
 
         GHashTable  *resources;
+                        
+        GSource     *timeout_src;
+        guint        num_discovery;
 };
 
 enum {
@@ -77,7 +83,7 @@ static guint signals[LAST_SIGNAL];
 typedef struct {
         GSSDPResourceBrowser *resource_browser;
         char                 *usn;
-        guint                 timeout_id;
+        GSource              *timeout_src;
 } Resource;
 
 /* Function prototypes */
@@ -87,13 +93,22 @@ gssdp_resource_browser_set_client (GSSDPResourceBrowser *resource_browser,
 static void
 message_received_cb              (GSSDPClient          *client,
                                   const char           *from_ip,
+                                  gushort               from_port,
                                   _GSSDPMessageType     type,
-                                  GHashTable           *headers,
+                                  SoupMessageHeaders   *headers,
                                   gpointer              user_data);
 static void
 resource_free                    (gpointer              data);
 static void
 clear_cache                      (GSSDPResourceBrowser *resource_browser);
+static void
+send_discovery_request            (GSSDPResourceBrowser *resource_browser);
+static gboolean
+discovery_timeout                (gpointer              data);
+static void
+start_discovery                  (GSSDPResourceBrowser *resource_browser);
+static void
+stop_discovery                   (GSSDPResourceBrowser *resource_browser);
 
 static void
 gssdp_resource_browser_init (GSSDPResourceBrowser *resource_browser)
@@ -197,7 +212,9 @@ gssdp_resource_browser_dispose (GObject *object)
                                 (resource_browser->priv->client,
                                  resource_browser->priv->message_received_id);
                 }
-                                                   
+
+                stop_discovery (resource_browser);
+
                 g_object_unref (resource_browser->priv->client);
                 resource_browser->priv->client = NULL;
         }
@@ -492,20 +509,12 @@ gssdp_resource_browser_set_active (GSSDPResourceBrowser *resource_browser,
         resource_browser->priv->active = active;
 
         if (active) {
-                /* Emit discovery message */
-                char *message;
+                start_discovery (resource_browser);
+        } else {
+                stop_discovery (resource_browser);
 
-                message = g_strdup_printf (SSDP_DISCOVERY_REQUEST,
-                                           resource_browser->priv->target,
-                                           resource_browser->priv->mx);
-
-                _gssdp_client_send_message (resource_browser->priv->client,
-                                            NULL,
-                                            message);
-
-                g_free (message);
-        } else
                 clear_cache (resource_browser);
+        }
         
         g_object_notify (G_OBJECT (resource_browser), "active");
 }
@@ -547,25 +556,25 @@ resource_expire (gpointer user_data)
 
 static void
 resource_available (GSSDPResourceBrowser *resource_browser,
-                    GHashTable           *headers)
+                    SoupMessageHeaders   *headers)
 {
-        GSList *list;
         const char *usn;
+        const char *header;
         Resource *resource;
         gboolean was_cached;
         guint timeout;
         GList *locations;
+        GMainContext *context;
 
-        list = g_hash_table_lookup (headers, "USN");
-        if (!list)
+        usn = soup_message_headers_get (headers, "USN");
+        if (!usn)
                 return; /* No USN specified */
-        usn = list->data;
 
         /* Get from cache, if possible */
         resource = g_hash_table_lookup (resource_browser->priv->resources, usn);
         if (resource) {
                 /* Remove old timeout */
-                g_source_remove (resource->timeout_id);
+                g_source_destroy (resource->timeout_src);
 
                 was_cached = TRUE;
         } else {
@@ -583,15 +592,17 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         }
 
         /* Calculate new timeout */
-        list = g_hash_table_lookup (headers, "Cache-Control");
-        if (list) {
-                GSList *l;
+        header = soup_message_headers_get (headers, "Cache-Control");
+        if (header) {
+                GSList *list;
                 int res;
 
                 res = 0;
 
-                for (l = list; l; l = l->next) {
-                        res = sscanf (l->data,
+                for (list = soup_header_parse_list (header);
+                     list;
+                     list = list->next) {
+                        res = sscanf (list->data,
                                       "max-age = %d",
                                       &timeout);
                         if (res == 1)
@@ -603,16 +614,24 @@ resource_available (GSSDPResourceBrowser *resource_browser,
                                    "default max-age of %d.\n"
                                    "Header was:\n%s",
                                    SSDP_DEFAULT_MAX_AGE,
-                                   (char *) list->data);
+                                   header);
 
                         timeout = SSDP_DEFAULT_MAX_AGE;
                 }
+
+                soup_header_free_list (list);
         } else {
-                list = g_hash_table_lookup (headers, "Expires");
-                if (list) {
+                const char *expires;
+
+                expires = soup_message_headers_get (headers, "Expires");
+                if (expires) {
+                        SoupDate *soup_exp_time;
                         time_t exp_time, cur_time;
 
-                        exp_time = soup_date_parse (list->data);
+                        soup_exp_time = soup_date_new_from_string (expires);
+                        exp_time = soup_date_to_time_t (soup_exp_time);
+                        soup_date_free (soup_exp_time);
+
                         cur_time = time (NULL);
 
                         if (exp_time > cur_time)
@@ -622,7 +641,7 @@ resource_available (GSSDPResourceBrowser *resource_browser,
                                            "default max-age of %d.\n"
                                            "Header was:\n%s",
                                            SSDP_DEFAULT_MAX_AGE,
-                                           (char *) list->data);
+                                           expires);
 
                                 timeout = SSDP_DEFAULT_MAX_AGE;
                         }
@@ -635,9 +654,16 @@ resource_available (GSSDPResourceBrowser *resource_browser,
                 }
         }
 
-        resource->timeout_id = g_timeout_add (timeout * 1000,
-                                              resource_expire,
-                                              resource);
+        resource->timeout_src = g_timeout_source_new_seconds (timeout);
+	g_source_set_callback (resource->timeout_src,
+                               resource_expire,
+			       resource, NULL);
+
+        context = gssdp_client_get_main_context
+                (resource_browser->priv->client);
+	g_source_attach (resource->timeout_src, context);
+
+        g_source_unref (resource->timeout_src);
 
         /* Only continue with signal emission if this resource was not
          * cached already */
@@ -647,17 +673,18 @@ resource_available (GSSDPResourceBrowser *resource_browser,
         /* Build list of locations */
         locations = NULL;
 
-        list = g_hash_table_lookup (headers, "Location");
-        if (list)
-                locations = g_list_append (locations, g_strdup (list->data));
+        header = soup_message_headers_get (headers, "Location");
+        if (header)
+                locations = g_list_append (locations, g_strdup (header));
 
-        list = g_hash_table_lookup (headers, "AL");
-        if (list) {
+        header = soup_message_headers_get (headers, "AL");
+        if (header) {
                 /* Parse AL header. The format is:
                  * <uri1><uri2>... */
-                char *start, *end, *uri;
+                const char *start, *end;
+                char *uri;
                 
-                start = list->data;
+                start = header;
                 while ((start = strchr (start, '<'))) {
                         start += 1;
                         if (!start || !*start)
@@ -691,15 +718,13 @@ resource_available (GSSDPResourceBrowser *resource_browser,
 
 static void
 resource_unavailable (GSSDPResourceBrowser *resource_browser,
-                      GHashTable           *headers)
+                      SoupMessageHeaders   *headers)
 {
-        GSList *list;
         const char *usn;
 
-        list = g_hash_table_lookup (headers, "USN");
-        if (!list)
+        usn = soup_message_headers_get (headers, "USN");
+        if (!usn)
                 return; /* No USN specified */
-        usn = list->data;
 
         /* Only process if we were cached */
         if (!g_hash_table_lookup (resource_browser->priv->resources, usn))
@@ -715,16 +740,16 @@ resource_unavailable (GSSDPResourceBrowser *resource_browser,
 
 static void
 received_discovery_response (GSSDPResourceBrowser *resource_browser,
-                             GHashTable           *headers)
+                             SoupMessageHeaders   *headers)
 {
-        GSList *list;
+        const char *st;
 
-        list = g_hash_table_lookup (headers, "ST");
-        if (!list)
+        st = soup_message_headers_get (headers, "ST");
+        if (!st)
                 return; /* No target specified */
 
         if (strcmp (resource_browser->priv->target, GSSDP_ALL_RESOURCES) != 0 &&
-            strcmp (resource_browser->priv->target, list->data) != 0)
+            strcmp (resource_browser->priv->target, st) != 0)
                 return; /* Target doesn't match */
 
         resource_available (resource_browser, headers);
@@ -732,28 +757,28 @@ received_discovery_response (GSSDPResourceBrowser *resource_browser,
 
 static void
 received_announcement (GSSDPResourceBrowser *resource_browser,
-                       GHashTable           *headers)
+                       SoupMessageHeaders   *headers)
 {
-        GSList *list;
+        const char *header;
 
-        list = g_hash_table_lookup (headers, "NT");
-        if (!list)
+        header = soup_message_headers_get (headers, "NT");
+        if (!header)
                 return; /* No target specified */
 
         if (strcmp (resource_browser->priv->target, GSSDP_ALL_RESOURCES) != 0 &&
-            strcmp (resource_browser->priv->target, list->data) != 0)
+            strcmp (resource_browser->priv->target, header) != 0)
                 return; /* Target doesn't match */
 
-        list = g_hash_table_lookup (headers, "NTS");
-        if (!list)
+        header = soup_message_headers_get (headers, "NTS");
+        if (!header)
                 return; /* No announcement type specified */
 
         /* Check announcement type */
-        if      (strncmp (list->data,
+        if      (strncmp (header,
                           SSDP_ALIVE_NTS,
                           strlen (SSDP_ALIVE_NTS)) == 0)
                 resource_available (resource_browser, headers);
-        else if (strncmp (list->data,
+        else if (strncmp (header,
                           SSDP_BYEBYE_NTS,
                           strlen (SSDP_BYEBYE_NTS)) == 0)
                 resource_unavailable (resource_browser, headers);
@@ -763,11 +788,12 @@ received_announcement (GSSDPResourceBrowser *resource_browser,
  * Received a message
  **/
 static void
-message_received_cb (GSSDPClient      *client,
-                     const char       *from_ip,
-                     _GSSDPMessageType type,
-                     GHashTable       *headers,
-                     gpointer          user_data)
+message_received_cb (GSSDPClient        *client,
+                     const char         *from_ip,
+                     gushort             from_port,
+                     _GSSDPMessageType   type,
+                     SoupMessageHeaders *headers,
+                     gpointer            user_data)
 {
         GSSDPResourceBrowser *resource_browser;
 
@@ -800,7 +826,7 @@ resource_free (gpointer data)
 
         g_free (resource->usn);
 
-        g_source_remove (resource->timeout_id);
+        g_source_destroy (resource->timeout_src);
 
         g_slice_free (Resource, resource);
 }
@@ -831,3 +857,77 @@ clear_cache (GSSDPResourceBrowser *resource_browser)
                                      clear_cache_helper,
                                      NULL);
 }
+
+/* Sends discovery request */
+static void
+send_discovery_request (GSSDPResourceBrowser *resource_browser)
+{
+        char *message;
+
+        message = g_strdup_printf (SSDP_DISCOVERY_REQUEST,
+                                   resource_browser->priv->target,
+                                   resource_browser->priv->mx,
+                                   g_get_application_name () ?: "");
+
+        _gssdp_client_send_message (resource_browser->priv->client,
+                                    NULL,
+                                    0,
+                                    message);
+
+        g_free (message);
+}
+
+static gboolean
+discovery_timeout (gpointer data)
+{
+        GSSDPResourceBrowser *resource_browser;
+
+        resource_browser = GSSDP_RESOURCE_BROWSER (data);
+
+        send_discovery_request (resource_browser);
+
+        resource_browser->priv->num_discovery += 1;
+
+        if (resource_browser->priv->num_discovery >= MAX_DISCOVERY_MESSAGES) {
+		resource_browser->priv->timeout_src = NULL;
+                resource_browser->priv->num_discovery = 0;
+
+                return FALSE;
+        } else
+                return TRUE;
+}
+
+/* Starts sending discovery requests */
+static void
+start_discovery (GSSDPResourceBrowser *resource_browser)
+{
+        GMainContext *context;
+
+        /* Send one now */
+        send_discovery_request (resource_browser);
+
+        /* And schedule the rest for later */
+        resource_browser->priv->num_discovery = 1;
+        resource_browser->priv->timeout_src =
+                g_timeout_source_new (DISCOVERY_FREQUENCY);
+	g_source_set_callback (resource_browser->priv->timeout_src,
+			       discovery_timeout,
+			       resource_browser, NULL);
+
+        context = gssdp_client_get_main_context
+                (resource_browser->priv->client);
+	g_source_attach (resource_browser->priv->timeout_src, context);
+
+        g_source_unref (resource_browser->priv->timeout_src);
+}
+
+/* Stops the sending of discovery messages */
+static void
+stop_discovery (GSSDPResourceBrowser *resource_browser)
+{
+	if (resource_browser->priv->timeout_src) {
+		g_source_destroy (resource_browser->priv->timeout_src);
+		resource_browser->priv->timeout_src = NULL;
+		resource_browser->priv->num_discovery = 0;
+	}
+}
